# -*- coding: utf-8 -*-
# Copyright (C) 2018-2022 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""fd_peak.modules.unstack"""

import copy
import logging
import math
import threading
import time

from six import iteritems, itervalues
import requests

from dwho.classes.modules import DWhoModuleBase, MODULES
from httpdis.ext.httpdis_json import HttpReqErrJson
from sonicprobe.libs.moresynchro import RWLock
from sonicprobe.libs.workerpool import WorkerPool

import pycurl


LOG                                = logging.getLogger('fd_peak.modules.unstack')
QLIST_INTERVAL                     = 60
QSIZE_INTERVAL                     = 1
UNSTACK_INTERVAL                   = 1
UNSTACK_MAX_LOOP                   = 10000
UNSTACK_THRESHOLD_WORKERS_EXPECTED = 1000
MAX_BULK_SIZE                      = 10

logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('urllib3').setLevel(logging.WARNING)


class QueueList(threading.Thread):
    def __init__(self, conf, result):
        threading.Thread.__init__(self)
        self.conf     = conf
        self.daemon   = True
        self.name     = 'qlist'
        self.killed   = False
        self.result   = result
        self.interval = float(conf.get('interval') or QLIST_INTERVAL)
        if self.interval < 0:
            self.interval = QLIST_INTERVAL

    def kill(self):
        self.killed = True

    def run(self):
        while not self.killed:
            interval = self.interval
            req      = None

            try:
                req = requests.get(self.conf['url'],
                                   params = {'site_id': '0'})
                if req and req.text:
                    qlist = req.json()
                    if qlist:
                        self.result['qlist'] = qlist
                    else:
                        self.result['qlist'] = None
                        interval = 1
            except Exception as e:
                LOG.exception(e)
            finally:
                if req:
                    req.close()
            time.sleep(interval)


class QueueSize(threading.Thread):
    def __init__(self, conf, result):
        threading.Thread.__init__(self)
        self.conf     = conf
        self.daemon   = True
        self.name     = 'qsize'
        self.killed   = False
        self.result   = result
        self.interval = float(conf.get('interval') or QSIZE_INTERVAL)
        if self.interval < 0:
            self.interval = QSIZE_INTERVAL

    def kill(self):
        self.killed = True

    def run(self):
        while not self.killed:
            if not self.result['qlist']:
                self.result['qsize'] = {}
                time.sleep(self.interval)
                continue

            qlist = list(self.result['qlist'])
            for q in qlist:
                if self.killed:
                    break

                req = None

                try:
                    req = requests.get(self.conf['url'],
                                       params = {'site_id': q})
                    if req and req.text:
                        self.result['qsize'][q] = req.json()
                except Exception as e:
                    LOG.exception(e)
                finally:
                    if req:
                        req.close()
            time.sleep(self.interval)


class QueueUnstack(threading.Thread):
    def __init__(self, conf, result):
        threading.Thread.__init__(self)
        self.conf       = copy.copy(conf)
        self.daemon     = True
        self.name       = 'unstack'
        self.killed     = False
        self.result     = result
        self.workers    = {}

        self.wconf      = conf.get('worker') or {}
        self.wconf['queue'] = None

        self.interval   = float(conf.get('interval') or UNSTACK_INTERVAL)
        if self.interval < 0:
            self.interval = UNSTACK_INTERVAL

        self.threshold_workers_expected = int(conf.get('threshold_workers_expected', UNSTACK_THRESHOLD_WORKERS_EXPECTED))

        max_loop        = self.conf.get('max_loop') or UNSTACK_MAX_LOOP
        self.unstack_range = range(0, max_loop)
        if not self.unstack_range:
            raise ValueError("invalid unstack.max_loop: %r" % max_loop)

        self.max_bulk_size = float(conf.get('max_bulk_size') or MAX_BULK_SIZE)
        if self.max_bulk_size < 1:
            raise ValueError("invalid unstack.max_bulk_size: %r" % self.max_bulk_size)

    def kill(self):
        self.killed = True

        for worker in itervalues(self.workers):
            worker.killall(1)

    def unstack(self, **kargs):
        for x in self.unstack_range: # pylint: disable-msg=unused-variable
            if self.killed:
                break

            req = None

            try:
                req = pycurl.Curl() # pylint: disable-msg=c-extension-no-member
                req.setopt(req.URL, "%s?site_id=%s" % (self.conf['url'], kargs['qname']))
                req.perform()
                if req.getinfo(req.RESPONSE_CODE) != 200:
                    break
            except Exception as e:
                LOG.exception(e)
            finally:
                if req:
                    req.close()
            time.sleep(0.05)

    def run(self):
        while not self.killed:
            qsize = copy.copy(self.result['qsize'])
            if not qsize:
                time.sleep(self.interval)
                continue

            for qname, size in iteritems(qsize):
                if self.killed:
                    break

                if size <= 0:
                    if qname in self.workers and self.workers[qname].tasks.empty():
                        self.workers[qname].killall(5)
                        del self.workers[qname]
                    time.sleep(self.interval)
                    continue

                if qname not in self.workers:
                    wconf               = copy.copy(self.wconf)
                    wconf['name']       = qname
                    self.workers[qname] = WorkerPool(**wconf)

                w = self.workers[qname]

                nb_unstack = int(math.ceil(float(size) / self.max_bulk_size))
                if nb_unstack <= 0:
                    time.sleep(self.interval)
                    continue

                total_working    = w.count_working()
                total_workers    = w.workers
                total_expected   = 0
                expected_working = nb_unstack - total_working
                if expected_working > 0:
                    total_expected = expected_working
                LOG.info("queue => name: %r, size: %r, workers => working: %r, waiting: %r, expected: %r (%r)",
                          qname,
                          size,
                          total_working,
                          total_workers - total_working,
                          total_expected,
                          expected_working)
                if expected_working >= 0:
                    if total_workers > 0 and total_workers >= self.threshold_workers_expected:
                        w.add(total_workers / self.threshold_workers_expected)
                    w.run(self.unstack, **{'qname': qname})
                time.sleep(self.interval)


class UnstackModule(DWhoModuleBase):
    MODULE_NAME = 'unstack'

    LOCK        = RWLock()

    # pylint: disable-msg=attribute-defined-outside-init
    def safe_init(self, options):
        self.lock_timeout = self.config['general']['lock_timeout']

        self.result       = {'qlist': None,
                             'qsize': {}}

        self.tqlist       = QueueList(self.modconf['qlist'], self.result)
        self.tqsize       = QueueSize(self.modconf['qsize'], self.result)
        self.tunstack     = QueueUnstack(self.modconf['unstack'], self.result)

    def at_start(self, options):
        self.tqlist.start()
        self.tqsize.start()
        self.tunstack.start()

    def at_stop(self):
        self.tqlist.kill()
        self.tqsize.kill()
        self.tunstack.kill()

    def status(self, request): # pylint: disable-msg=unused-variable,unused-argument
        if not self.LOCK.acquire_read(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for reading after %s seconds" % self.lock_timeout)

        try:
            return self.result
        except HttpReqErrJson as e:
            raise
        except Exception as e:
            LOG.exception(e)
            raise HttpReqErrJson(500)
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(UnstackModule())
    _start()
